/*____________________________________________________________________________
	
	Ahonto - TFolderTreeView
 
	Copyright (C) 2001 Project Ahonto
 
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
 
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
 
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Project Ahonto
 
____________________________________________________________________________*/

#include <vcl.h>
#pragma hdrstop

#include "FolderTreeView.h"
#include "ItemID.h"
#include "SHNotify.h"
#pragma package(smart_init)

//---------------------------------------------------------------------------
// ValidCtrCheck ́A`ꂽR|[lgz֐܂
// ۃNXł͂ȂƂmF邽߂ɒ`Ă܂B
//

__fastcall TFolderTreeNode::TFolderTreeNode(TTreeNodes* AOwner)
        : TTreeNode(AOwner)
{
    FItemID      = NULL;
    FItemIDList  = NULL;
    FIncludeHidden = ((TFolderTreeView*)AOwner->Owner)->ShowHidden;
}

__fastcall TFolderTreeNode::~TFolderTreeNode()
{
    ReleaseItemID();
}

void __fastcall TFolderTreeNode::SetShellFolder(_di_IShellFolder _ShellFolder)
{
    FShellFolder = _ShellFolder;
}

void __fastcall TFolderTreeNode::SetItemID(PItemIDList ItemID)
{
    TFolderTreeNode *ParentNode;
    PItemIDList ParentItemID = NULL;

    ReleaseItemID();

    FItemID = ItemID;

    if (Parent)
    {
        ParentItemID = ((TFolderTreeNode*)Parent)->ItemIDList;
    }
    FItemIDList = ConcatItemID(ParentItemID, ItemID);
}

PItemIDList __fastcall TFolderTreeNode::GetItemIDList(void)
{
    return FItemIDList;
}

void __fastcall TFolderTreeNode::ReleaseItemID(void)
{
    if (FItemID)
        CoTaskMemFree(FItemID);
    if (FItemIDList)
        CoTaskMemFree(FItemIDList);
}

AnsiString __fastcall TFolderTreeNode::GetPath(void)
{
    char Path[MAX_PATH];

    SHGetPathFromIDList(ItemIDList, Path);
    return AnsiString(Path);
}

void __fastcall TFolderTreeNode::MakeSubFolder()
{
    _di_IEnumIDList  EnumIDList;
    TFolderTreeNode* CurrentNode = (TFolderTreeNode*)getFirstChild();

    try
    {
        Owner->BeginUpdate();

        int EnumAttr = SHCONTF_FOLDERS;

        if (FIncludeHidden)
            EnumAttr |= SHCONTF_INCLUDEHIDDEN;

        if (S_OK == FShellFolder->EnumObjects(NULL, EnumAttr, &EnumIDList))
        {
            PItemIDList SubItemID;

            while (S_OK == EnumIDList->Next(1, &SubItemID, NULL))
            {
                ULONG Attribute = SFGAO_CAPABILITYMASK
                                  | SFGAO_HASSUBFOLDER
                                  | SFGAO_FILESYSTEM
                                  | SFGAO_FILESYSANCESTOR
                                  | SFGAO_SHARE;

                FShellFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&SubItemID, &Attribute);

                if (Attribute & SFGAO_FILESYSTEM || Attribute & SFGAO_FILESYSANCESTOR)
                {
                    _di_IShellFolder    SubFolder;
                    PItemIDList         ItemIDList;
                    TFolderTreeNode*    SubNode;
                    TFolderTreeNode*    TmpNode = (TFolderTreeNode*)GetNextChild(CurrentNode);
                    AnsiString          DisplayName = GetDisplayName(FShellFolder, SubItemID, false);

                    FShellFolder->BindToObject(SubItemID, NULL, IID_IShellFolder, (void**)&SubFolder);

                    if (CurrentNode == NULL)
                    {
                        SubNode = (TFolderTreeNode*)Owner->AddChild(this, DisplayName);
                    }
                    else
                    {
                        if (GetScode(FShellFolder->CompareIDs(0, SubItemID, CurrentNode->ItemID)))
                        {
                            CurrentNode->Delete();
                            CurrentNode = TmpNode;

                            if (CurrentNode != NULL)
                                SubNode = (TFolderTreeNode*)CurrentNode->Owner->Insert(CurrentNode, DisplayName);
                            else
                                SubNode = (TFolderTreeNode*)Owner->AddChild(this, DisplayName);

                        }
                        else
                        {
                            SubNode = CurrentNode;
                            CurrentNode = TmpNode;
                        }
                    }
                    SubNode->ShellFolder = SubFolder;
                    SubNode->ItemID = SubItemID;
                    SubNode->ImageIndex = GetShellImage(SubNode->ItemIDList, false, false);
                    SubNode->SelectedIndex = GetShellImage(SubNode->ItemIDList, false, true);
                    SubNode->HasChildren = Attribute & SFGAO_HASSUBFOLDER;

                    if (Attribute & SFGAO_SHARE)
                        SubNode->OverlayIndex = 0;
                }
            }

            while(CurrentNode)
            {
                TFolderTreeNode* tmpNode = (TFolderTreeNode*)GetNextChild(CurrentNode);
                CurrentNode->Delete();
                CurrentNode = tmpNode;
            }
        }
    }
    __finally
    {
        Owner->EndUpdate();
    }
}

__fastcall TFolderTreeView::TFolderTreeView(TComponent* AOwner)
        : TCustomTreeView(AOwner)
{}

void __fastcall TFolderTreeView::WndProc(Messages::TMessage &Message)
{
    POINT Point;
    bool handled = false;

    switch(Message.Msg)
    {
    case WM_CONTEXTMENU:
        Point.x = Message.LParamLo;
        Point.y = Message.LParamHi;
        ContextPopup((TPoint*)&Point, handled);
        Message.Result = handled;
        break;
    case WM_INITMENUPOPUP:
        ContextMenu2->HandleMenuMsg(Message.Msg, Message.WParam, Message.LParam);
        Message.Result = 0;
        break;
    case WM_DRAWITEM:
    case WM_MEASUREITEM:
        ContextMenu2->HandleMenuMsg(Message.Msg, Message.WParam, Message.LParam);
        Message.Result = 1;
        break;
    case WM_SHNOTIFY:
        LPSHNOTIFYSTRUCT shntfy = (LPSHNOTIFYSTRUCT)Message.WParam;

        for (int i = 0; i < Items->Count; i++)
        {
            PItemIDList ItemID;
            TFolderTreeNode* Node = (TFolderTreeNode*)Items->Item[i];

            if (shntfy->dwItem1)
            {
                ItemID = GetParentItemID((PItemIDList)shntfy->dwItem1);
                if (!GetScode(FIDesktopFolder->CompareIDs(0, ItemID, Node->ItemIDList)))
                {
                    Node->MakeSubFolder();
                }
                CoTaskMemFree(ItemID);
            }

            if (shntfy->dwItem2)
            {
                ItemID = GetParentItemID((PItemIDList)shntfy->dwItem2);
                if (!GetScode(FIDesktopFolder->CompareIDs(0, ItemID, Node->ItemIDList)))
                {
                    Node->MakeSubFolder();
                }
                CoTaskMemFree(ItemID);
            }
        }
        Message.Result = 1;
        break;
    }
    TCustomTreeView::WndProc(Message);
}

TTreeNode* __fastcall TFolderTreeView::CreateNode(void)
{
    return new TFolderTreeNode(Items);
}

void __fastcall TFolderTreeView::Loaded()
{
    // no implements
}

void __fastcall TFolderTreeView::CreateRootNode(PItemIDList RootItemID)
{
    TFolderTreeNode*  RootNode;
    NOTIFYREGISTER _register;

    RootNode = (TFolderTreeNode*)Items->Add(NULL, GetDisplayName(FIDesktopFolder, RootItemID, false));

    RootNode->ShellFolder = FIDesktopFolder;
    RootNode->ItemID = RootItemID;
    RootNode->ImageIndex     = GetShellImage(RootItemID, false, false);
    RootNode->SelectedIndex  = GetShellImage(RootItemID, false, true);
    RootNode->HasChildren    = true;

    RootNode->MakeSubFolder();
    RootNode->Expand(false);
}

bool __fastcall TFolderTreeView::CanExpand(TTreeNode* Node)
{
    TFolderTreeNode* FolderNode = (TFolderTreeNode*)Node;

    FolderNode->MakeSubFolder();
    FolderNode->MakeVisible();
    if (!FolderNode->Count)
        FolderNode->HasChildren = false;

    return TCustomTreeView::CanExpand(FolderNode);
}

void __fastcall TFolderTreeView::Change(TTreeNode *Node)
{
    if (FOnChange)
        FOnChange(this, (TFolderTreeNode*)Node);
}

void __fastcall TFolderTreeView::DragDrop(TObject* Source, int X, int Y)
{
    if (DropTarget && FOnDragDrop)
    {
        FOnDragDrop(this, Source, (TFolderTreeNode*)DropTarget, X, Y);
    }
}

void __fastcall TFolderTreeView::DragOver(System::TObject* Source, int X, int Y,
        TDragState State, bool &Accept)
{
    if (DropTarget && FOnDragOver)
    {
        FOnDragOver(this, Source, (TFolderTreeNode*)DropTarget, X, Y, State, Accept);
    }
}

void __fastcall TFolderTreeView::Edit(const tagTVITEMA &Item)
{
    HRESULT	hRes;
    OLECHAR	ochNewName[MAX_PATH];
    PItemIDList NewItemID;
    TFolderTreeNode* Node = (TFolderTreeNode*)Items->GetNode(Item.hItem);
    TFolderTreeNode* ParentNode = (TFolderTreeNode*)Node->Parent;

    if (Item.pszText && ParentNode)
    {

        ::MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, Item.pszText, -1,
                               ochNewName, MAX_PATH );

        hRes = ParentNode->ShellFolder->SetNameOf(Application->Handle,
                Node->ItemID,
                ochNewName
                ,SHCONTF_FOLDERS | SHCONTF_INCLUDEHIDDEN
                ,&NewItemID );

        if (hRes == NOERROR)
        {
            Node->ItemID = NewItemID;
            TCustomTreeView::Edit(Item);
        }
    }
}


void __fastcall TFolderTreeView::ContextPopup(Windows::TPoint* MousePos,
        bool &Handled)
{
    _di_IContextMenu  ContextMenu;
    HRESULT result;
    int                 idCmd;
    HMENU               hMenu;
    CMINVOKECOMMANDINFO cmi;

    TFolderTreeNode* Node = (TFolderTreeNode*)Selected;
    PItemIDList ItemID = Node->ItemID;
    _di_IShellFolder ShellFolder;

    if (!Node->Parent)
        return;

    ShellFolder = ((TFolderTreeNode*)Node->Parent)->ShellFolder;

    result = ShellFolder->GetUIObjectOf(Handle,
                                        1,
                                        (LPCITEMIDLIST*)&ItemID,
                                        IID_IContextMenu,
                                        0,
                                        (void**)&ContextMenu);

    if(FAILED(result))
        return;

    result = ContextMenu->QueryInterface(IID_IContextMenu2, (void **)&ContextMenu2);

    if(SUCCEEDED(result))
    {
        hMenu = CreatePopupMenu();

        if(hMenu)
        {
            result = ContextMenu2->QueryContextMenu(hMenu,
                                                    0,
                                                    1,
                                                    0x7fff,
                                                    CMF_NORMAL );
            if (SUCCEEDED(result))
            {
                idCmd = TrackPopupMenu(hMenu,
                                       TPM_LEFTALIGN | TPM_RETURNCMD | TPM_RIGHTBUTTON,
                                       MousePos->x,
                                       MousePos->y,
                                       0,
                                       Handle,
                                       NULL);
                if (idCmd)
                {
                    //Jt@C̃IvVƂ\TCYƂ̐ݒłB
                    cmi.cbSize = sizeof(CMINVOKECOMMANDINFO);
                    cmi.fMask  = 0;
                    cmi.hwnd   = Handle;
                    cmi.lpVerb = MAKEINTRESOURCE(idCmd-1);
                    cmi.lpParameters = NULL;
                    cmi.lpDirectory  = NULL;
                    cmi.nShow        = SW_SHOWNORMAL;
                    cmi.dwHotKey     = 0;
                    cmi.hIcon        = NULL;
                    //̃R}hŎs܂
                    ContextMenu2->InvokeCommand(&cmi);
                }
            }
            DestroyMenu(hMenu);
        }
        Handled = true;
    }
}

void __fastcall TFolderTreeView::ExpandByPath(AnsiString Path)
{
    OLECHAR olePath[MAX_PATH];
    PItemIDList OrgItemID;
    ULONG Attribute = 0;
    ULONG Length = MAX_PATH;

    MultiByteToWideChar( CP_ACP, 0, Path.c_str(), -1, olePath, MAX_PATH);

    OleCheck( FIDesktopFolder->ParseDisplayName( Application->Handle, NULL,
              olePath, &Length, &OrgItemID, &Attribute));

    PItemIDList nextedItemID = OrgItemID;
    PItemIDList ComparedItemID = CopyOneItemID(nextedItemID);

    try
    {
        Items->BeginUpdate();
        for (int i = 0; i < Items->Count; i++)
        {
            HRESULT result;
            PItemIDList ItemIDList = ((TFolderTreeNode*)Items->Item[i])->ItemIDList;

            result = FIDesktopFolder->CompareIDs(0, ItemIDList, ComparedItemID);

            if (FAILED(result))
                continue;

            if (SCODE_CODE(GetScode(result)) == 0)
            {
                Items->Item[i]->Selected = true;
                nextedItemID = NextItemID(nextedItemID);

                if (!nextedItemID->mkid.cb)
                    break;

                Items->Item[i]->Expand(false);
                PItemIDList TmpItemID = ConcatItemID(ComparedItemID, CopyOneItemID(nextedItemID));
                CoTaskMemFree(ComparedItemID);
                ComparedItemID = TmpItemID;
            }
        }
    }
    __finally
    {
        Items->EndUpdate();
        if (Selected)
        Selected->MakeVisible();
    }
    CoTaskMemFree(OrgItemID);
}

void __fastcall TFolderTreeView::SetIncludeHidden(bool Hidden)
{
    FIncludeHidden = Hidden;
}

AnsiString __fastcall TFolderTreeView::GetSelectedPath(void)
{
    if (Selected)
    {
        return ((TFolderTreeNode*)Selected)->Path;
    }
    return AnsiString();
}

void __fastcall TFolderTreeView::CreateWnd()
{
    TSHFileInfo Fileinfo;
    THandle     ImageListHandle;
    PItemIDList RootItemID;
    NOTIFYREGISTER _register;

    TCustomTreeView::CreateWnd();

    OleCheck(SHGetDesktopFolder( &FIDesktopFolder ));
    OleCheck(SHGetSpecialFolderLocation(Application->Handle, CSIDL_DESKTOP, &RootItemID));


    _register.pidlPath = RootItemID;
    _register.bWatchSubtree = true;

    hNotify = _SHChangeNotifyRegister(Handle,
                                      SHCNF_TYPE | SHCNF_IDLIST,
                                      SHCNE_RENAMEFOLDER | SHCNE_INTERRUPT | SHCNE_MKDIR
                                      //SHCNE_RENAMEFOLDER | SHCNE_MKDIR
                                      | SHCNE_RMDIR | SHCNE_UPDATEDIR,
                                      WM_SHNOTIFY,
                                      1,
                                      &_register);

    ImageListHandle = SHGetFileInfo("C:\\", NULL, &Fileinfo, sizeof(Fileinfo),
                                    SHGFI_SMALLICON | SHGFI_SYSICONINDEX);

    TreeView_SetImageList(Handle, ImageListHandle, TVSIL_NORMAL);

    CreateRootNode(RootItemID);
}

void __fastcall TFolderTreeView::DestroyWnd()
{
    if (hNotify)
        _SHChangeNotifyDeregister(hNotify);

    TCustomTreeView::DestroyWnd();
}

static inline void ValidCtrCheck(TFolderTreeView *)
{
    new TFolderTreeView((TComponent*)NULL);
}

namespace Foldertreeview
{
void __fastcall PACKAGE Register()
{
    TComponentClass classes[1] = {__classid(TFolderTreeView)};
    RegisterComponents("Sasaki", classes, 0);
}
}
