
#ifndef FOLDER_TREE_VIEW
#define FOLDER_TREE_VIEW

#include <Classes.hpp>
#include <Controls.hpp>
#include <StdCtrls.hpp>
#include <ComCtrls.hpp>
#include <ComObj.hpp>
#include "FindChange.h"

class TFolderTreeNode : public TTreeNode
{
private:
    _di_IShellFolder    FShellFolder;
    PItemIDList         FItemID;
    PItemIDList         FItemIDList;
    bool FIncludeHidden;
protected:
    void __fastcall SetShellFolder(_di_IShellFolder);
    void __fastcall SetItemID(PItemIDList ItemID);
    PItemIDList __fastcall GetItemIDList(void);
    AnsiString __fastcall GetPath(void);
    void __fastcall SetSearched(bool Search);
    void __fastcall ReleaseItemID(void);
public:
    __fastcall TFolderTreeNode(TTreeNodes* AOwner);
    __fastcall ~TFolderTreeNode();
    void __fastcall MakeSubFolder();

    __property _di_IShellFolder ShellFolder = {read = FShellFolder, write=SetShellFolder };
    __property PItemIDList ItemID = {read = FItemID, write=SetItemID };
    __property PItemIDList ItemIDList = {read = GetItemIDList};
    __property AnsiString Path = {read = GetPath};
};

typedef void __fastcall (__closure *TFTVChangedEvent)(System::TObject* Sender,
        TFolderTreeNode* Node);

typedef void __fastcall (__closure *TFTVDragDropEvent)(System::TObject* Sender,
        System::TObject* Source, TFolderTreeNode* Target, int X, int Y);

typedef void __fastcall (__closure *TFTVDragOverEvent)(System::TObject* Sender,
        System::TObject* Source, TFolderTreeNode* Target, int X, int Y, TDragState State, bool &Accept);

class PACKAGE TFolderTreeView : public TCustomTreeView
{
private:
    _di_IShellFolder FIDesktopFolder;
    _di_IContextMenu2 ContextMenu2;
    TFTVChangedEvent FOnChange;
    TFTVDragDropEvent FOnDragDrop;
    TFTVDragOverEvent FOnDragOver;
    bool FIncludeHidden;
    HANDLE hNotify;
protected:
    void __fastcall Loaded();
    TTreeNode* __fastcall CreateNode(void);
    void __fastcall CreateRootNode(PItemIDList RootItemID);
    DYNAMIC bool __fastcall CanExpand(TTreeNode* Node);
    DYNAMIC void __fastcall Change(TTreeNode *Node);
    DYNAMIC void __fastcall DragDrop(TObject* Source, int X, int Y);
    DYNAMIC void __fastcall DragOver(System::TObject* Source, int X, int Y,
                                     TDragState State, bool &Accept);
    DYNAMIC void __fastcall ContextPopup(Windows::TPoint* MousePos, bool &Handled);
    DYNAMIC void __fastcall Edit(const tagTVITEMA &Item);
    void __fastcall ExpandByPath(AnsiString Path);
    AnsiString __fastcall GetSelectedPath(void);
    void __fastcall SetIncludeHidden(bool Hidden);
    friend void __fastcall MakeNode(TFolderTreeNode* Node);
    void __fastcall CreateWnd();
    void __fastcall DestroyWnd();
    void __fastcall WndProc(Messages::TMessage &Message);
public:
    __fastcall TFolderTreeView(Classes::TComponent* AOwner);

__published:
    __property Align ;
    __property Anchors ;
    __property AutoExpand ;
    __property BiDiMode ;
    __property BorderStyle ;
    __property BorderWidth ;
    __property ChangeDelay ;
    __property Color ;
    __property Ctl3D ;
    __property Constraints ;
    __property DragKind ;
    __property DragCursor ;
    __property DragMode ;
    __property Enabled ;
    __property AnsiString FolderName = {read=GetSelectedPath, write=ExpandFromPath};
    __property Font ;
    __property HideSelection ;
    __property HotTrack ;
    //	__property Images ;
    __property Indent ;
    __property ParentBiDiMode ;
    __property ParentColor ;
    __property ParentCtl3D ;
    __property ParentFont ;
    __property ParentShowHint ;
    __property PopupMenu ;
    //	__property ReadOnly ;
    __property RightClickSelect ;
    __property RowSelect ;
    //__property ShowButtons ;
    __property ShowHint ;
    __property ShowLines ;
__property bool ShowHidden = {read =FIncludeHidden, write = SetIncludeHidden, default = false};
    //	__property ShowRoot ;
    //	__property SortType ;
    //	__property StateImages ;
    __property TabOrder ;
    __property TabStop ;
    __property ToolTips ;
    __property Visible ;
    //__property OnAdvancedCustomDraw ;
    //__property OnAdvancedCustomDrawItem ;
    __property TFTVChangedEvent OnChange = {read=FOnChange, write=FOnChange};
    __property OnChanging ;
    __property OnClick ;
    __property OnCollapsed ;
    __property OnCollapsing ;
    //__property OnCompare ;
    __property OnContextPopup ;
    //__property OnCustomDraw ;
    //__property OnCustomDrawItem ;
    __property OnDblClick ;
    __property OnDeletion ;
    __property TFTVDragDropEvent OnDragDrop = { read =FOnDragDrop, write=FOnDragDrop};
    __property TFTVDragOverEvent OnDragOver = { read =FOnDragOver, write=FOnDragOver};
    //__property OnEdited ;
    //__property OnEditing ;
    __property OnEndDock ;
    __property OnEndDrag ;
    __property OnEnter ;
    __property OnExit ;
    __property OnExpanding ;
    __property OnExpanded ;
    //__property OnGetImageIndex ;
    //__property OnGetSelectedIndex ;
    __property OnKeyDown ;
    __property OnKeyPress ;
    __property OnKeyUp ;
    __property OnMouseDown ;
    __property OnMouseMove ;
    __property OnMouseUp ;
    __property OnStartDock ;
    __property OnStartDrag ;
    //	__property Items ;
}
;

#endif

