#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>

#include "MBString.h"
#pragma package(smart_init)

union _MULTIBYTES
{
    struct
    {
        unsigned char high;
        unsigned char lo;
    };
    unsigned short word;
};

MBString::MBString (MBString & _code)
{
    length = _code.length;
    charcode = _code.charcode;
    SetLength(strlen(_code.data));
    strcpy(data, _code.data);
}

MBString::~MBString ()
{
    SetLength (0);
}

MBString::MBString (char* _s)
{
    CharSet code = (CharSet)CodeDetect(_s);

    Init(_s, code);
}

MBString::MBString (char* _s, CharSet _code)
{
    Init(_s, _code);
}

void MBString::Init (char* _s, CharSet _code)
{
    length = strlen(_s);
    this->data = NULL;

    switch (_code)
    {
    case ISO_2022_JP:
        SetLength (GetCRLFCount(_s) + 1);
        JISToSJIS (data, _s);
        charcode = ISO_2022_JP;
        break;
    case EUC_JP:
        SetLength (GetCRLFCount(_s) + 1);
        EUCToSJIS (data, _s);
        charcode = EUC_JP;
        break;
    default:
        SetLength(GetCRLFCount(_s) + 1);
        LFToCRLF(data, _s);
        charcode = _code;
    }
}

int MBString::Size ()
{
    return length;
}

const char* MBString::ToStr()
{
    return data;
}

void MBString::SetLength (int _alloc_len)
{
    data = (char *) realloc (data, _alloc_len);
}

int MBString::GetCRLFCount (char *_str)
{
    int length = 0;
    int kcode = 0;
    unsigned char *buf = _str;

    while (*buf)
    {
        length++;
        if (*buf++ == '\n')
            kcode++;
    }
    return length + (kcode * 2);
}

inline bool IsBin (unsigned char *code)
{
    if (*code <= 0x07 || *code == 0xFF)
        return true;
    return false;
}

inline bool IsJIS (unsigned char *code)
{
    if (*code++ == 0x1B && *code == '$')
        return true;
    return false;
}

inline bool IsEUC (unsigned char *code)
{
    if (*code++ > 0xA0 && *code > 0xA0)
        return true;
    return false;
}

inline bool IsSJIS (unsigned char *code)
{
    if (((*code > 0x80 && *code < 0xA0) || (*code > 0xDF && *code < 0xFD))
            &&*++code > 0x3F && *code != 0x7F)
        return true;
    return false;
}

int MBString::CodeDetect (unsigned char *_s)
{
    int sjis = 0;
    int euc = 0;
    int asc = 0;

    while (*_s && *(_s + 1))
    {
        if (IsBin (_s))
            return BINARY;
        else if (IsJIS (_s))
            return ISO_2022_JP;
        else if (IsEUC (_s))
            euc++;
        else if (IsSJIS (_s))
            sjis++;
        else
            asc++;
        _s++;
    }
    if (sjis > euc)
        return SHIFT_JIS;
    else if (euc)
        return EUC_JP;
    else if (asc)
        return USASCII;
    return UNKNOWN;
}

bool null_terminated (char *buf, int length)
{
    if (length == 1)
        if (*buf)
            return false;
    if (length == 2)
        if (*buf++)
            if (*buf)
                return false;
    if (length == 3)
        if (*buf++)
            if (*buf++)
                if (*buf)
                    return false;
    if (length == 4)
        if (*buf++)
            if (*buf++)
                if (*buf++)
                    if (*buf)
                        return false;
    return true;
}


inline _MULTIBYTES* inline_jis2sjis (_MULTIBYTES *mb)
{
    //    if (mb->high & 0xFF) return 0;

    //    if (mb->high >= 0x21 && mb->high <= 0x7e && mb->lo >= 0x21 && mb->lo <= 0x7e)
    //    {
    if (mb->high & 0x01)
    {
        mb->high >>= 1;
        mb->high < 0x2f ? mb->high += 0x71 :  mb->high -= 0x4f;
        mb->lo += 0x1f;
        if (mb->lo >= 0x7f)
            mb->lo++;
    }
    else
    {
        mb->high >>= 1;
        mb->high < 0x2f ? mb->high += 0x70 :  mb->high -= 0x50;
        mb->lo += 0x7e;
    }

    return mb;
}

inline _MULTIBYTES* inline_euc2sjis (_MULTIBYTES *mb)
{
    mb->word ^= (short)0x8080;
    inline_jis2sjis (mb);
    return mb;
}

char* MBString::JISToSJIS (unsigned char *_dest, unsigned char *_str)
{
    _MULTIBYTES mb;

    if (*_str && *_str == '\n')
    {
        *_dest++ = '\r';
        *_dest++ = *_str++;
    }

    while (*_str)
    {
        if (IsJIS (_str) && !null_terminated (_str, 3))
        {
            _str += 3;
            while (*_str != 0x1B && !null_terminated (_str, 2))
            {
                mb.high = _str[0];
                mb.lo   = _str[1];
                inline_jis2sjis (&mb);
                *_dest++ = mb.high;
                *_dest++ = mb.lo;
                _str += 2;
            }
            _str += 3;
        }
        else if (*_str == '\n')
        {
            *_dest++ = '\r';
            *_dest++ = *_str++;
        }
        else
        {
            *_dest++ = *_str++;
        }
    }
    *_dest = '\0';
    return _dest;
}

char *MBString::EUCToSJIS (unsigned char *_dest, unsigned char *_str)
{
    _MULTIBYTES mb;

    if (*_str && *_str == '\n')
    {
        *_dest++ = '\r';
        *_dest++ = *_str++;
    }

    while (*_str)
    {
        if (IsEUC (_str) && !null_terminated (_str, 2))
        {
            mb.word = *((unsigned short*)(_str));
            inline_euc2sjis (&mb);
            *((unsigned short*)_dest)++ = mb.word;
            _str += 2;
        }
        else if (*_str == '\n')
        {
            *_dest++ = '\r';
            *_dest++ = *_str++;
        }
        else
        {
            *_dest++ = *_str++;
        }
    }
    *_dest = '\0';
    return _dest;
}

char *MBString::LFToCRLF (unsigned char *_dest, unsigned char *_str)
{
    _MULTIBYTES mb;

    if (*_str && *_str == '\n')
    {
        *_dest++ = '\r';
        *_dest++ = *_str++;
    }

    while (*_str)
    {
        if (*_str == '\n' && _str[-1] != '\r')
        {
            *_dest++ = '\r';
            *_dest++ = *_str++;
        }
        else
        {
            *_dest++ = *_str++;
        }
    }
    *_dest = '\0';
    return _dest;
}

