
#include <windows.h>

#define WM_SHNOTIFY	WM_USER + 100

#define SHCNF_ACCEPT_INTERRUPTS 0x0001
#define SHCNF_ACCEPT_NON_INTERRUPTS 0x0002
#define SHCNF_NO_PROXY 0x8000

typedef struct _NOTIFYREGISTER
{
    LPCITEMIDLIST  pidlPath;
    BOOL           bWatchSubtree;
}
NOTIFYREGISTER, *LPNOTIFYREGISTER;

typedef const LPNOTIFYREGISTER LPCNOTIFYREGISTER;

struct _DWORDITEMID
{
    USHORT cb;
    DWORD  dwItem1;
    DWORD  dwItem2;
}
DWORDITEMID, *LPDWORDITEMID;

typedef struct tagSHNOTIFYSTRUCT
{
    DWORD dwItem1;
    DWORD dwItem2;

}
SHNOTIFYSTRUCT, *LPSHNOTIFYSTRUCT;

#ifdef __cplusplus
extern "C"
{
#endif

    HANDLE
    WINAPI
    SHChangeNotifyRegister(
        HWND   hWnd,
        DWORD  dwFlags,
        LONG   wEventMask,
        UINT   uMsg,
        DWORD  cItems,
        LPCNOTIFYREGISTER lpItems
    );


    // hWnd        - Handle of the window to receive the window message specified in uMsg.

    // uFlags        - Flag that indicates the meaning of the dwItem1 and dwItem2 members of the
    //                     SHNOTIFYSTRUCT (which is pointed to by the window procedure//s wParam
    //                     value when the specifed window message is received). This parameter can
    //                     be one of the SHCN_ItemFlags enum values below.
    //                     This interpretaion may be inaccurate as it appears pidls are almost alway returned
    //                     in the SHNOTIFYSTRUCT despite this value. See James// site for more info...

    // dwEventId - Combination of SHCN_EventIDs enum values that specifies what events the
    //                     specified window will be notified of. See below.

    // uMsg          - Window message to be used to identify receipt of a shell change notification.
    //                      The message should *not* be a value that lies within the specifed window//s
    //                      message range ( i.e. BM_ messages for a button window) or that window may
    //                      not receive all (if not any) notifications sent by the shell!!!

    // cItems         - Count of PIDLSTRUCT structures in the array pointed to by the lpps param.

    // lpps             - Pointer to an array of PIDLSTRUCT structures indicating what folder(s) to monitor
    //                      changes in, and whether to watch the specified folder//s subfolder.

    // If successful, returns a notification handle which must be passed to SHChangeNotifyDeregister
    // when no longer used. Returns 0 otherwise.

    // Once the specified message is registered with SHChangeNotifyRegister, the specified
    // window//s function proc will be notified by the shell of the specified event in (and under)
    // the folder(s) speciifed in apidl. On message receipt, wParam points to a SHNOTIFYSTRUCT
    // and lParam contains the event//s ID value.

    // The values in dwItem1 and dwItem2 are event specific. See the description of the values
    // for the wEventId parameter of the documented SHChangeNotify API function.

    // Closes the notification handle returned from a call to SHChangeNotifyRegister.
    // Returns True if succeful, False otherwise.

    BOOL
    WINAPI
    SHChangeNotifyDeregister(
        HANDLE hNotify
    );

    BOOL
    WINAPI
    SHChangeNotification_Unlock(
        HANDLE hLock
    );

    HANDLE
    WINAPI
    SHChangeNotification_Lock(
        HANDLE  hMemoryMap,
        DWORD   dwProcessId,
        LPCITEMIDLIST **lppidls,
        LPLONG  lpwEventId
    );

    HANDLE (WINAPI *_SHChangeNotifyRegister)(HWND, DWORD, LONG, UINT, DWORD, LPCNOTIFYREGISTER);
    HANDLE (WINAPI *_SHChangeNotifyDeregister)(HANDLE);
    HANDLE (WINAPI *_SHChangeNotification_Lock)(HANDLE, DWORD, LPCITEMIDLIST, LPLONG);
    HANDLE (WINAPI *_SHChangeNotification_Unlock)(HANDLE);

#ifdef __cplusplus
}
#endif

struct SHChangeNotifyLoad
{
    SHChangeNotifyLoad()
    {
        HANDLE hLib = GetModuleHandle("SHELL32");

        (FARPROC&)_SHChangeNotifyRegister = GetProcAddress(hLib, MAKEINTRESOURCE(2));
        (FARPROC&)_SHChangeNotifyDeregister = GetProcAddress(hLib, MAKEINTRESOURCE(4));
        (FARPROC&)_SHChangeNotification_Lock = GetProcAddress(hLib, MAKEINTRESOURCE(644));
        (FARPROC&)_SHChangeNotification_Unlock = GetProcAddress(hLib, MAKEINTRESOURCE(645));
    }
};

SHChangeNotifyLoad Dummy;

