#include <stdio.h>
#include <stdlib.h>

#ifdef __WIN32__
#include <windows.h>
#include <winsock.h>
#else
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#endif

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#include "URLParser.h"

//---------------------------------------------------------------------------

bool URLParser::parse(char* _URL)
{
    char *serviceNameDelimiter;
    servent *service;
    char *pathDelimiter;
    char *portDelimiter;

    serviceNameDelimiter = strstr(_URL, "://");
    if (serviceNameDelimiter)
    {
        *serviceNameDelimiter = '\0';
        strcpy(serviceName, _URL);
        _URL = _URL + (serviceNameDelimiter - _URL) + 3;

        service = getservbyname(serviceName, "tcp");

        if (!service)
            return false;
        port = ntohs(service->s_port);
    }

    pathDelimiter = strstr(_URL, "/");
    if (pathDelimiter != 0)
    {
        strcpy(path, pathDelimiter);
        *pathDelimiter = '\0';
        strcpy(hostName, _URL);
    }
    else if (strlen(_URL) < sizeof(hostName))
    {
        strcpy(hostName, _URL);
        strcpy(path, "/");
    }
    else
    {
        return false;
    }

    portDelimiter = strstr(hostName, ":");
    if (portDelimiter != 0)
    {
        *portDelimiter = '\0';
        port = atoi(portDelimiter + 1);
        if (port == 0)
        {
            return false;
        }
    }
    else if(!serviceNameDelimiter)
    {
        service = getservbyport(htons(port), "tcp");
        if (!service)
        {
            return false;
        }
        strcpy(serviceName, service->s_name);
    }
    return true;
}

URLParser::URLParser(AnsiString URL, char* defaultProto)
{
    servent *service = getservbyname(defaultProto, "tcp");
    if (!service)
    {
        throw URLParserException();
    }
    port = ntohs(service->s_port);

    if (!parse(URL.c_str()))
    {
        throw URLParserException();
    }
}

URLParser::URLParser(char* URL, char* defaultProto)
{
#if defined( __WIN32__ ) && defined (SINGLES)
    InitSocket();
#endif

    servent *service = getservbyname(defaultProto, "tcp");
    if (!service)
    {
        throw URLParserException();
    }
    port = ntohs(service->s_port);

    if (!parse(URL))
    {
        throw URLParserException();
    }
}

URLParser::URLParser(URLParser &_URL)
{
    memcpy(this, &_URL, sizeof(_URL));
}

URLParser::~URLParser()
{
#if defined( __WIN32__ ) && defined (SINGLES)
    CleanupSocket();
#endif
}

char* URLParser::getServiceName()
{
    return serviceName;
}

char* URLParser::getHostName()
{
    return hostName;
}

unsigned short URLParser::getPort()
{
    return port;
}

char* URLParser::getPath()
{
    return path;
}

char* URLParser::getURL()
{
    servent *service = getservbyname(serviceName, "tcp");
    if (!service)
        return NULL;

    if (port == ntohs(service->s_port))
    {
        sprintf(realURL, "%s://%s%s", serviceName, hostName, path);
    }
    else
    {
        sprintf(realURL, "%s://%s:%d/%s", serviceName, hostName, port, path);
    }
    return realURL;
}
